import datetime

from pyconomic import base
from pyconomic import models


class VatZone:
    Domestic = "HomeCountry"
    EU = "EU"
    Abroad = "Abroad"


class EconomicsFacade(object):
    def __init__(self, client, service):
        self.client = client
        self.service = service

    def _build_types(self):
        modelcode = base.build_model_code(self.client)
        exec modelcode in globals()
        return modelcode

    def initialize(self, url, agreement_number, user_id, password):
        conf = {
            'agreementNumber': str(agreement_number),
            'userName': user_id,
            'password': password
        }
        self.service.Connect(**conf)

    def get_all_changes(self):
        return self.service.get_all_changes()

    def commit(self):
        self.service.commit()

    def find(self, model, **spec):
        return self.service.find(model, **spec)

    def get(self, model, **spec):
        return self.service.get(model, **spec)

    def get_or_create(self, model, **spec):
        """
        Args:
            model: class of Model
            get_data: the filter used for finding an instance
            create_data: the data used to create an instance, if none could be found
        """
        return self.service.get_or_create(model, **spec)

    def create(self, model, **data):
        return self.service.create(model, **data)

    def num_server_calls(self):
        return self.service.ncalls

    def send_order(self, order):
        self.service.fetch('Order_RegisterAsSent', order._handle)

    def create_order(self, debtor, is_vat_included=True, due_date=None,
                     heading='', text_line1='', text_line2='',
                     debtor_data=None, delivery_data=None, products=None,
                     project=None, other_reference='', model=models.Order, **extra
                     ):
        """Create a new Order.

        Args:
            debtor (Debtor): the debtor of the order
            debtor_data (mapping): map of debtor data {'postal_code: .., 'city': .., 'ean': ..}
                                   defaults to values on debitor instance for missing values
            delivery_data (mapping): map of delivery data {'address': ..., 'postal_code': ...}
                                     defaults to values on debitor instance for missing values
            due_date (datetime): due date
            heading (string): heading to be displayed in the order pdf
            text_line1 (string): first order description line
            text_line2 (string): second order description line
            other_reference (string): custom string to be used for identification
            extra (mapping): mapping of extra values to be passed in to the server call
        Returns:
            Order instance
        """

        debtor_data = debtor_data or {}
        delivery_data = delivery_data or {}
        delivery_date = delivery_data.get('date', datetime.datetime.now())
        our_reference = extra.get('our_reference', debtor.our_reference)
        currency = extra.get('currency', debtor.currency)
        layout = extra.get('layout', debtor.layout)
        term_of_payment = extra.get('term_of_payment', debtor.term_of_payment)
        date = extra.get('date', datetime.datetime.now())
        order_input = {
            'debtor': debtor,
            'number': extra.get('number', 1),
            'project': project,
        }
        for dd in ['name', 'address', 'postal_code', 'city', 'country', 'ean']:
            order_input['debtor_%s' % dd] = debtor_data.get(dd, getattr(debtor, dd))

        for dd in ['address', 'postal_code', 'city', 'country']:
            order_input['delivery_%s' % dd] = delivery_data.get(dd, getattr(debtor, dd))

        order_input.update({
            'delivery_date': delivery_date or datetime.datetime.now(),
            'heading': heading,
            'text_line1': text_line1,
            'text_line2': text_line2,
            'is_archived': extra.get('is_archived', 0),
            'is_sent': extra.get('is_sent', 0),
            'net_amount': extra.get('net_amount', 0),
            'vat_amount': extra.get('vat_amount', 0),
            'gross_amount': extra.get('gross_amount', 0),
            'margin': extra.get('margin', 0),
            'margin_as_percent': extra.get('margin_as_percent', 0),
            'date': date,
            'our_reference': our_reference,
            'other_reference': other_reference,
            'currency': currency,
            'exchange_rate': extra.get('exchange_rate', 1.0),
            'is_vat_included': is_vat_included,
            'layout': layout,
            'due_date': due_date or datetime.datetime.now(),
            'term_of_payment': term_of_payment
        })
        order_input.update(extra)
        order = self.create(model, **order_input)
        if products:
            for product in products:
                self.create_orderline(order, product)
        return order

    def create_orderline(self, order, product, quantity=1, description=None, number=1,
                         unit_net_price=None, unit_cost_price=None, total_net_amount=None, discount_as_percent=0,
                         delivery_date=None, total_margin=0, margin_as_percent=0, model=models.OrderLine
                         ):
        delivery_date = delivery_date or datetime.datetime.now()
        unit_net_price = unit_net_price or product.sales_price
        unit_cost_price = unit_cost_price or product.cost_price
        total_net_amount = total_net_amount or product.sales_price
        description = product.description if description is None else description

        line_input = {
            'order': order,
            'product': product,
            'number': number,
            'description': description,
            'delivery_date': delivery_date,
            'quantity': quantity,
            'unit_net_price': unit_net_price,
            'unit_cost_price': unit_cost_price,
            'total_net_amount': total_net_amount,
            'discount_as_percent': discount_as_percent,
            'total_margin': total_margin,
            'margin_as_percent': margin_as_percent
        }
        return self.create(model, **line_input)

    def create_product(self, number, product_group, name, sales_price, description='', cost_price=0.0,
                       in_stock=100.0, recommended_price=0.0, distribution_key=None,
                       volume=100.0, department=None, available=0.0, is_accessible=True,
                       unit=None, bar_code='', on_order=0.0, model=models.Product
                       ):

        product_input = {
            'distribution_key': distribution_key,
            'department': department,
            'product_group': product_group,
            'unit': unit,

            'number': number,

            'on_order': on_order,
            'name': name,
            'sales_price': sales_price,
            'description': description,
            'cost_price': cost_price,
            'in_stock': in_stock,
            'recommended_price': recommended_price,
            'volume': volume,
            'available': available if available is not None else in_stock,
            'bar_code': bar_code,
            'is_accessible': is_accessible
        }
        return self.create(model, **product_input)

    def create_debtor(self, debtor_group, name, term_of_payment, vat_zone, currency,
                      number=None, is_accessible=True, model=models.Debtor, **extra):
        if number is None:
            number = self.service.next_available_number(model)
        debtor_input = {
            'name': name,
            'number': number,
            'debtor_group': debtor_group,
            'term_of_payment': term_of_payment,
            'vat_zone': vat_zone,
            'currency': currency,
            'is_accessible': is_accessible
        }
        debtor_input.update(extra)
        return self.create(model, **debtor_input)

    def upgrade_to_invoice(self, order, model=models.CurrentInvoice):
        return self.service.upgrade_to_invoice(order._handle, model)

    def upgrade_to_order(self, quotation, model=models.Order):
        return self.service.upgrade_to_order(quotation._handle, model)

    def book_invoice(self, current_invoice, model=models.Invoice):
        return self.service.book_invoice(current_invoice._handle, model)

    def find_orders(self):
        pass

    def find_invoices(self):
        pass

    def find_accounts(self):
        pass


def Pyconomic(agreement_number, user_id, password):
    url = 'https://api.e-conomic.com/secure/api1/EconomicWebservice.asmx?WSDL'
    model_factory = base.EConomicsModelFactory()
    wsdl = base.Client(url)
    codec = base.PropertyCodec()
    service = base.EConomicsService(wsdl.service, model_factory, wsdl.factory, codec)
    
    facade = EconomicsFacade(wsdl, service)
    facade.initialize(
        url,
        agreement_number, user_id, password
    )
    return facade


from pyconomic.base import *


class AccountingPeriod(EConomicsModel):
    __filters__ = [{'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    from_date = EConomicsProperty('FromDate')
    status = EConomicsProperty('Status')
    to_date = EConomicsProperty('ToDate')
    period = EConomicsProperty('Period')
    accounting_year = EConomicsReference('AccountingYear', 'AccountingYear')


class CostTypeGroup(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}]
    account_closed = EConomicsProperty('AccountClosed')
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')
    account_ongoing = EConomicsProperty('AccountOngoing')


class DebtorEntry(EConomicsModel):
    __filters__ = [{'name': u'invoice_number', 'method': u'FindByInvoiceNumber'}, {'name': u'serial_number', 'method': u'FindBySerialNumber'}]
    invoice_number = EConomicsProperty('InvoiceNumber')
    remainder = EConomicsProperty('Remainder')
    due_date = EConomicsProperty('DueDate')
    amount = EConomicsProperty('Amount')
    serial_number = EConomicsProperty('SerialNumber')
    currency = EConomicsReference('Currency', 'Currency')
    type = EConomicsProperty('Type')
    text = EConomicsProperty('Text')
    account = EConomicsReference('Account', 'Account')
    date = EConomicsProperty('Date')
    amount_default_currency = EConomicsProperty('AmountDefaultCurrency')
    remainder_default_currency = EConomicsProperty('RemainderDefaultCurrency')
    debtor = EConomicsReference('Debtor', 'Debtor')
    voucher_number = EConomicsProperty('VoucherNumber')


class Creditor(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    number = EConomicsProperty('Number')
    our_reference = EConomicsReference('OurReference', 'Employee')
    city = EConomicsProperty('City')
    term_of_payment = EConomicsReference('TermOfPayment', 'TermOfPayment')
    entries = EConomicsReferenceList('Entries', 'Entry', 'Creditor_GetEntries')
    default_payment_type = EConomicsProperty('DefaultPaymentType')
    creditor_contacts = EConomicsReferenceList('CreditorContacts', 'CreditorContact', 'Creditor_GetCreditorContacts')
    creditor_group = EConomicsReference('CreditorGroup', 'CreditorGroup')
    currency = EConomicsReference('Currency', 'Currency')
    ci_number = EConomicsProperty('CINumber')
    is_accessible = EConomicsProperty('IsAccessible')
    vat_zone = EConomicsProperty('VatZone')
    open_entries = EConomicsProperty('OpenEntries')
    default_payment_creditor_id = EConomicsProperty('DefaultPaymentCreditorId')
    name = EConomicsProperty('Name')
    entries_by_invoice_no = EConomicsProperty('EntriesByInvoiceNo')
    address = EConomicsProperty('Address')
    email = EConomicsProperty('Email')
    entries_by_voucher_no_and_invoice_no = EConomicsProperty('EntriesByVoucherNoAndInvoiceNo')
    bank_account = EConomicsProperty('BankAccount')
    auto_contra_account = EConomicsProperty('AutoContraAccount')
    entries_by_voucher_no = EConomicsProperty('EntriesByVoucherNo')
    country = EConomicsProperty('Country')
    county = EConomicsProperty('County')
    your_reference = EConomicsProperty('YourReference')
    attention = EConomicsProperty('Attention')
    postal_code = EConomicsProperty('PostalCode')


class ExtendedVatZone(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}]
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class CostType(EConomicsModel):
    __filters__ = [{'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'name', 'method': u'FindByName'}]
    vat_account = EConomicsReference('VatAccount', 'VatAccount')
    number = EConomicsProperty('Number')
    name = EConomicsProperty('Name')
    cost_type_group = EConomicsReference('CostTypeGroup', 'CostTypeGroup')


class TemplateCollection(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}]
    name = EConomicsProperty('Name')


class BudgetFigure(EConomicsModel):
    __filters__ = [{'name': u'date', 'method': u'FindByDate'}]
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    text = EConomicsProperty('Text')
    to_date = EConomicsProperty('ToDate')
    account = EConomicsReference('Account', 'Account')
    from_date = EConomicsProperty('FromDate')
    amount_default_currency = EConomicsProperty('AmountDefaultCurrency')
    department = EConomicsReference('Department', 'Department')


class DistributionKey(EConomicsModel):
    __filters__ = [{'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}]
    number = EConomicsProperty('Number')
    percentages = EConomicsProperty('Percentages')
    departments = EConomicsReferenceList('Departments', 'Department', 'DistributionKey_GetDepartments')
    name = EConomicsProperty('Name')
    is_accessible = EConomicsProperty('IsAccessible')


class ReportCodeSet(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}]
    code_set = EConomicsProperty('CodeSet')
    name = EConomicsProperty('Name')
    codes = EConomicsProperty('Codes')


class ProductGroup(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'name', 'method': u'FindByName'}]
    accrual = EConomicsProperty('Accrual')
    name = EConomicsProperty('Name')
    account_for_vat_exempt_debtor_invoices_current = EConomicsProperty('AccountForVatExemptDebtorInvoicesCurrent')
    products = EConomicsReferenceList('Products', 'Product', 'ProductGroup_GetProducts')
    account_for_vat_liable_debtor_invoices_current = EConomicsProperty('AccountForVatLiableDebtorInvoicesCurrent')
    number = EConomicsProperty('Number')


class Subscriber(EConomicsModel):
    __filters__ = [{'name': u'subscripton_list', 'method': u'FindBySubscriptonList'}, {'name': u'subscription', 'method': u'FindBySubscription'}]
    quantity_factor = EConomicsProperty('QuantityFactor')
    debtor = EConomicsReference('Debtor', 'Debtor')
    project = EConomicsReference('Project', 'Project')
    price_index = EConomicsProperty('PriceIndex')
    extra_text_for_invoice = EConomicsProperty('ExtraTextForInvoice')
    special_price = EConomicsProperty('SpecialPrice')
    start_date = EConomicsProperty('StartDate')
    expiry_date = EConomicsProperty('ExpiryDate')
    discount_as_percent = EConomicsProperty('DiscountAsPercent')
    discount_expiry_date = EConomicsProperty('DiscountExpiryDate')
    registered_date = EConomicsProperty('RegisteredDate')
    comments = EConomicsProperty('Comments')
    end_date = EConomicsProperty('EndDate')
    subscription = EConomicsReference('Subscription', 'Subscription')


class Department(EConomicsModel):
    __filters__ = [{'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class CreditorEntry(EConomicsModel):
    __filters__ = [{'name': u'invoice_number', 'method': u'FindByInvoiceNumber'}, {'name': u'serial_number', 'method': u'FindBySerialNumber'}, {'name': u'serial_number_type_and_invoice_number', 'method': u'FindBySerialNumberTypeAndInvoiceNumber'}]
    remainder_default_currency = EConomicsProperty('RemainderDefaultCurrency')
    date = EConomicsProperty('Date')
    amount = EConomicsProperty('Amount')
    voucher_number = EConomicsProperty('VoucherNumber')
    account = EConomicsReference('Account', 'Account')
    type = EConomicsProperty('Type')
    serial_number = EConomicsProperty('SerialNumber')
    remainder = EConomicsProperty('Remainder')
    invoice_number = EConomicsProperty('InvoiceNumber')
    currency = EConomicsReference('Currency', 'Currency')
    due_date = EConomicsProperty('DueDate')
    amount_default_currency = EConomicsProperty('AmountDefaultCurrency')
    text = EConomicsProperty('Text')
    creditor = EConomicsReference('Creditor', 'Creditor')


class AccountingYear(EConomicsModel):
    __filters__ = [{'name': u'date', 'method': u'FindByDate'}]
    is_closed = EConomicsProperty('IsClosed')
    periods = EConomicsProperty('Periods')
    year = EConomicsProperty('Year')
    from_date = EConomicsProperty('FromDate')
    to_date = EConomicsProperty('ToDate')


class ReportCode(EConomicsModel):
    __filters__ = []
    parent_code = EConomicsProperty('ParentCode')
    code = EConomicsProperty('Code')
    accounts = EConomicsReferenceList('Accounts', 'Account', 'ReportCode_GetAccounts')
    full_code = EConomicsProperty('FullCode')


class Employee(EConomicsModel):
    __filters__ = [{'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}]
    name = EConomicsProperty('Name')
    mileage_entries_by_date = EConomicsProperty('MileageEntriesByDate')
    cost_price = EConomicsProperty('CostPrice')
    group = EConomicsProperty('Group')
    sales_price = EConomicsProperty('SalesPrice')
    type = EConomicsProperty('Type')
    time_entries_by_date = EConomicsProperty('TimeEntriesByDate')
    time_entries = EConomicsReferenceList('TimeEntries', 'TimeEntry', 'Employee_GetTimeEntries')
    number = EConomicsProperty('Number')


class OrderLine(EConomicsModel):
    __filters__ = [{'name': u'product', 'method': u'FindByProduct'}, {'name': u'product_list', 'method': u'FindByProductList'}]
    quantity = EConomicsProperty('Quantity')
    discount_as_percent = EConomicsProperty('DiscountAsPercent')
    description = EConomicsProperty('Description')
    order = EConomicsReference('Order', 'Order')
    department = EConomicsReference('Department', 'Department')
    unit_net_price = EConomicsProperty('UnitNetPrice')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    unit = EConomicsReference('Unit', 'Unit')
    margin_as_percent = EConomicsProperty('MarginAsPercent')
    product = EConomicsReference('Product', 'Product')
    number = EConomicsProperty('Number')
    unit_cost_price = EConomicsProperty('UnitCostPrice')
    delivery_date = EConomicsProperty('DeliveryDate')
    total_net_amount = EConomicsProperty('TotalNetAmount')
    total_margin = EConomicsProperty('TotalMargin')


class EmployeeGroup(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'number_list', 'method': u'FindByNumberList'}]
    employees = EConomicsReferenceList('Employees', 'Employee', 'EmployeeGroup_GetEmployees')
    number = EConomicsProperty('Number')
    name = EConomicsProperty('Name')


class CurrentInvoice(EConomicsModel):
    __filters__ = [{'name': u'our_reference', 'method': u'FindByOurReference'}, {'name': u'other_reference', 'method': u'FindByOtherReference'}, {'name': u'date_interval', 'method': u'FindByDateInterval'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    delivery_location = EConomicsReference('DeliveryLocation', 'DeliveryLocation')
    delivery_country = EConomicsProperty('DeliveryCountry')
    is_vat_included = EConomicsProperty('IsVatIncluded')
    debtor_city = EConomicsProperty('DebtorCity')
    debtor_postal_code = EConomicsProperty('DebtorPostalCode')
    net_amount = EConomicsProperty('NetAmount')
    lines = EConomicsReferenceList('Lines', 'CurrentInvoiceLine', 'CurrentInvoice_GetLines')
    deduction_amount = EConomicsProperty('DeductionAmount')
    margin_as_percent = EConomicsProperty('MarginAsPercent')
    vat_amount = EConomicsProperty('VatAmount')
    currency = EConomicsReference('Currency', 'Currency')
    margin = EConomicsProperty('Margin')
    delivery_city = EConomicsProperty('DeliveryCity')
    date = EConomicsProperty('Date')
    debtor = EConomicsReference('Debtor', 'Debtor')
    delivery_county = EConomicsProperty('DeliveryCounty')
    other_reference = EConomicsProperty('OtherReference')
    debtor_county = EConomicsProperty('DebtorCounty')
    delivery_address = EConomicsProperty('DeliveryAddress')
    public_entry_number = EConomicsProperty('PublicEntryNumber')
    debtor_ean = EConomicsProperty('DebtorEan')
    pdf = EConomicsFileProperty('Pdf', 'CurrentInvoice_GetPdf', 'pdf')
    layout = EConomicsReference('Layout', 'TemplateCollection')
    terms_of_delivery = EConomicsProperty('TermsOfDelivery')
    due_date = EConomicsProperty('DueDate')
    our_reference2 = EConomicsReference('OurReference2', 'Employee')
    your_reference = EConomicsProperty('YourReference')
    our_reference = EConomicsReference('OurReference', 'Employee')
    gross_amount = EConomicsProperty('GrossAmount')
    exchange_rate = EConomicsProperty('ExchangeRate')
    delivery_date = EConomicsProperty('DeliveryDate')
    rounding_amount = EConomicsProperty('RoundingAmount')
    attention = EConomicsProperty('Attention')
    delivery_postal_code = EConomicsProperty('DeliveryPostalCode')
    heading = EConomicsProperty('Heading')
    debtor_name = EConomicsProperty('DebtorName')
    debtor_address = EConomicsProperty('DebtorAddress')
    project = EConomicsReference('Project', 'Project')
    debtor_country = EConomicsProperty('DebtorCountry')
    term_of_payment = EConomicsReference('TermOfPayment', 'TermOfPayment')
    text_line1 = EConomicsProperty('TextLine1')
    text_line2 = EConomicsProperty('TextLine2')


class CurrentSupplierInvoiceLine(EConomicsModel):
    __filters__ = []
    unit_price = EConomicsProperty('UnitPrice')
    product = EConomicsReference('Product', 'Product')
    quantity = EConomicsProperty('Quantity')
    invoice = EConomicsReference('Invoice', 'Invoice')


class Product(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'bar_code', 'method': u'FindByBarCode'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'all_accessible', 'method': u'GetAllAccessible'}]
    on_order = EConomicsProperty('OnOrder')
    description = EConomicsProperty('Description')
    sales_price = EConomicsProperty('SalesPrice')
    in_stock = EConomicsProperty('InStock')
    recommended_price = EConomicsProperty('RecommendedPrice')
    cost_price = EConomicsProperty('CostPrice')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    number = EConomicsProperty('Number')
    volume = EConomicsProperty('Volume')
    name = EConomicsProperty('Name')
    department = EConomicsReference('Department', 'Department')
    available = EConomicsProperty('Available')
    ordered = EConomicsReference('Ordered', 'Order')
    product_group = EConomicsReference('ProductGroup', 'ProductGroup')
    unit = EConomicsReference('Unit', 'Unit')
    is_accessible = EConomicsProperty('IsAccessible')
    bar_code = EConomicsProperty('BarCode')


class ProductPrice(EConomicsModel):
    __filters__ = [{'name': u'product_and_currency', 'method': u'FindByProductAndCurrency'}]
    product = EConomicsReference('Product', 'Product')
    currency = EConomicsReference('Currency', 'Currency')
    price = EConomicsProperty('Price')


class Currency(EConomicsModel):
    __filters__ = [{'name': u'code', 'method': u'FindByCode'}]
    code = EConomicsProperty('Code')


class Company(EConomicsModel):
    __filters__ = []
    number = EConomicsProperty('Number')
    fax_number = EConomicsProperty('FaxNumber')
    county = EConomicsProperty('County')
    country = EConomicsProperty('Country')
    base_currency = EConomicsProperty('BaseCurrency')
    address2 = EConomicsProperty('Address2')
    address1 = EConomicsProperty('Address1')
    ci_number = EConomicsProperty('CINumber')
    city = EConomicsProperty('City')
    web_site = EConomicsProperty('WebSite')
    contact = EConomicsProperty('Contact')
    email = EConomicsProperty('Email')
    telephone_number = EConomicsProperty('TelephoneNumber')
    mobile_number = EConomicsProperty('MobileNumber')
    postal_code = EConomicsProperty('PostalCode')
    name = EConomicsProperty('Name')
    vat_number = EConomicsProperty('VatNumber')
    sign_up_date = EConomicsProperty('SignUpDate')


class CashBookEntry(EConomicsModel):
    __filters__ = []
    currency = EConomicsReference('Currency', 'Currency')
    amount = EConomicsProperty('Amount')
    creditor_invoice_number = EConomicsProperty('CreditorInvoiceNumber')
    debtor = EConomicsReference('Debtor', 'Debtor')
    vat_account = EConomicsReference('VatAccount', 'VatAccount')
    cost_type = EConomicsReference('CostType', 'CostType')
    cash_book = EConomicsReference('CashBook', 'CashBook')
    bank_payment_type = EConomicsReference('BankPaymentType', 'BankPaymentType')
    project = EConomicsReference('Project', 'Project')
    debtor_invoice_number = EConomicsProperty('DebtorInvoiceNumber')
    creditor = EConomicsReference('Creditor', 'Creditor')
    bank_payment_creditor_invoice_id = EConomicsProperty('BankPaymentCreditorInvoiceId')
    voucher_number = EConomicsProperty('VoucherNumber')
    type = EConomicsProperty('Type')
    contra_account = EConomicsProperty('ContraAccount')
    employee = EConomicsReference('Employee', 'Employee')
    text = EConomicsProperty('Text')
    account = EConomicsReference('Account', 'Account')
    end_date = EConomicsProperty('EndDate')
    due_date = EConomicsProperty('DueDate')
    amount_default_currency = EConomicsProperty('AmountDefaultCurrency')
    start_date = EConomicsProperty('StartDate')
    contra_vat_account = EConomicsProperty('ContraVatAccount')
    bank_payment_creditor_id = EConomicsProperty('BankPaymentCreditorId')
    department = EConomicsReference('Department', 'Department')
    date = EConomicsProperty('Date')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    capitalise = EConomicsProperty('Capitalise')


class Order(EConomicsModel):
    __filters__ = [{'name': u'all_current', 'method': u'GetAllCurrent'}, {'name': u'other_reference', 'method': u'FindByOtherReference'}, {'name': u'number_interval', 'method': u'FindByNumberInterval'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'date_interval', 'method': u'FindByDateInterval'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    other_reference = EConomicsProperty('OtherReference')
    gross_amount = EConomicsProperty('GrossAmount')
    public_entry_number = EConomicsProperty('PublicEntryNumber')
    net_amount = EConomicsProperty('NetAmount')
    margin_as_percent = EConomicsProperty('MarginAsPercent')
    debtor_county = EConomicsProperty('DebtorCounty')
    delivery_county = EConomicsProperty('DeliveryCounty')
    is_vat_included = EConomicsProperty('IsVatIncluded')
    lines = EConomicsReferenceList('Lines', 'OrderLine', 'Order_GetLines')
    project = EConomicsReference('Project', 'Project')
    pdf = EConomicsFileProperty('Pdf', 'Order_GetPdf', 'pdf')
    due_date = EConomicsProperty('DueDate')
    rounding_amount = EConomicsProperty('RoundingAmount')
    terms_of_delivery = EConomicsProperty('TermsOfDelivery')
    debtor_ean = EConomicsProperty('DebtorEan')
    layout = EConomicsReference('Layout', 'TemplateCollection')
    debtor_city = EConomicsProperty('DebtorCity')
    term_of_payment = EConomicsReference('TermOfPayment', 'TermOfPayment')
    vat_amount = EConomicsProperty('VatAmount')
    debtor_name = EConomicsProperty('DebtorName')
    debtor_country = EConomicsProperty('DebtorCountry')
    is_sent = EConomicsProperty('IsSent')
    debtor = EConomicsReference('Debtor', 'Debtor')
    delivery_date = EConomicsProperty('DeliveryDate')
    our_reference = EConomicsReference('OurReference', 'Employee')
    debtor_postal_code = EConomicsProperty('DebtorPostalCode')
    our_reference2 = EConomicsReference('OurReference2', 'Employee')
    debtor_address = EConomicsProperty('DebtorAddress')
    margin = EConomicsProperty('Margin')
    exchange_rate = EConomicsProperty('ExchangeRate')
    date = EConomicsProperty('Date')
    delivery_country = EConomicsProperty('DeliveryCountry')
    number = EConomicsProperty('Number')
    your_reference = EConomicsProperty('YourReference')
    text_line1 = EConomicsProperty('TextLine1')
    text_line2 = EConomicsProperty('TextLine2')
    delivery_address = EConomicsProperty('DeliveryAddress')
    is_archived = EConomicsProperty('IsArchived')
    delivery_city = EConomicsProperty('DeliveryCity')
    heading = EConomicsProperty('Heading')
    delivery_postal_code = EConomicsProperty('DeliveryPostalCode')
    delivery_location = EConomicsReference('DeliveryLocation', 'DeliveryLocation')
    attention = EConomicsProperty('Attention')
    currency = EConomicsReference('Currency', 'Currency')


class CurrentSupplierInvoice(EConomicsModel):
    __filters__ = []
    id = EConomicsProperty('Id')
    lines = EConomicsProperty('Lines')
    creditor = EConomicsReference('Creditor', 'Creditor')


class BankPaymentType(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'number_list', 'method': u'FindByNumberList'}]
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class DebtorGroup(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'number_list', 'method': u'FindByNumberList'}]
    name = EConomicsProperty('Name')
    account = EConomicsReference('Account', 'Account')
    debtors = EConomicsReferenceList('Debtors', 'Debtor', 'DebtorGroup_GetDebtors')
    layout = EConomicsReference('Layout', 'TemplateCollection')
    number = EConomicsProperty('Number')


class InvoiceLine(EConomicsModel):
    __filters__ = [{'name': u'invoice_number_interval', 'method': u'FindByInvoiceNumberInterval'}, {'name': u'product', 'method': u'FindByProduct'}, {'name': u'product_list', 'method': u'FindByProductList'}]
    product = EConomicsReference('Product', 'Product')
    number = EConomicsProperty('Number')
    discount_as_percent = EConomicsProperty('DiscountAsPercent')
    department = EConomicsReference('Department', 'Department')
    invoice = EConomicsReference('Invoice', 'Invoice')
    unit = EConomicsReference('Unit', 'Unit')
    quantity = EConomicsProperty('Quantity')
    vat_amount = EConomicsProperty('VatAmount')
    total_net_amount = EConomicsProperty('TotalNetAmount')
    description = EConomicsProperty('Description')
    vat_rate = EConomicsProperty('VatRate')
    delivery_date = EConomicsProperty('DeliveryDate')
    unit_cost_price = EConomicsProperty('UnitCostPrice')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    unit_net_price = EConomicsProperty('UnitNetPrice')


class Quotation(EConomicsModel):
    __filters__ = [{'name': u'date_interval', 'method': u'FindByDateInterval'}, {'name': u'other_reference', 'method': u'FindByOtherReference'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_current', 'method': u'GetAllCurrent'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'number_interval', 'method': u'FindByNumberInterval'}]
    debtor_address = EConomicsProperty('DebtorAddress')
    is_vat_included = EConomicsProperty('IsVatIncluded')
    public_entry_number = EConomicsProperty('PublicEntryNumber')
    date = EConomicsProperty('Date')
    currency = EConomicsReference('Currency', 'Currency')
    debtor_city = EConomicsProperty('DebtorCity')
    debtor_country = EConomicsProperty('DebtorCountry')
    due_date = EConomicsProperty('DueDate')
    debtor = EConomicsReference('Debtor', 'Debtor')
    margin = EConomicsProperty('Margin')
    other_reference = EConomicsProperty('OtherReference')
    our_reference = EConomicsReference('OurReference', 'Employee')
    delivery_address = EConomicsProperty('DeliveryAddress')
    your_reference = EConomicsProperty('YourReference')
    heading = EConomicsProperty('Heading')
    text_line2 = EConomicsProperty('TextLine2')
    text_line1 = EConomicsProperty('TextLine1')
    gross_amount = EConomicsProperty('GrossAmount')
    margin_as_percent = EConomicsProperty('MarginAsPercent')
    delivery_county = EConomicsProperty('DeliveryCounty')
    is_sent = EConomicsProperty('IsSent')
    delivery_date = EConomicsProperty('DeliveryDate')
    delivery_postal_code = EConomicsProperty('DeliveryPostalCode')
    terms_of_delivery = EConomicsProperty('TermsOfDelivery')
    our_reference2 = EConomicsReference('OurReference2', 'Employee')
    term_of_payment = EConomicsReference('TermOfPayment', 'TermOfPayment')
    debtor_name = EConomicsProperty('DebtorName')
    lines = EConomicsReferenceList('Lines', 'QuotationLine', 'Quotation_GetLines')
    delivery_location = EConomicsReference('DeliveryLocation', 'DeliveryLocation')
    layout = EConomicsReference('Layout', 'TemplateCollection')
    attention = EConomicsProperty('Attention')
    is_archived = EConomicsProperty('IsArchived')
    debtor_postal_code = EConomicsProperty('DebtorPostalCode')
    rounding_amount = EConomicsProperty('RoundingAmount')
    net_amount = EConomicsProperty('NetAmount')
    vat_amount = EConomicsProperty('VatAmount')
    debtor_county = EConomicsProperty('DebtorCounty')
    pdf = EConomicsProperty('Pdf')
    number = EConomicsProperty('Number')
    debtor_ean = EConomicsProperty('DebtorEan')
    delivery_city = EConomicsProperty('DeliveryCity')
    exchange_rate = EConomicsProperty('ExchangeRate')
    delivery_country = EConomicsProperty('DeliveryCountry')


class Debtor(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'ci_number', 'method': u'FindByCINumber'}, {'name': u'partial_name', 'method': u'FindByPartialName'}, {'name': u'telephone_and_fax_number', 'method': u'FindByTelephoneAndFaxNumber'}, {'name': u'ean', 'method': u'FindByEan'}, {'name': u'email', 'method': u'FindByEmail'}, {'name': u'number', 'method': u'FindByNumber'}]
    attention = EConomicsProperty('Attention')
    number = EConomicsProperty('Number')
    address = EConomicsProperty('Address')
    postal_code = EConomicsProperty('PostalCode')
    website = EConomicsProperty('Website')
    telephone_and_fax_number = EConomicsProperty('TelephoneAndFaxNumber')
    city = EConomicsProperty('City')
    your_reference = EConomicsProperty('YourReference')
    country = EConomicsProperty('Country')
    invoices = EConomicsReferenceList('Invoices', 'Invoice', 'Debtor_GetInvoices')
    extended_vat_zone = EConomicsReference('ExtendedVatZone', 'ExtendedVatZone')
    is_accessible = EConomicsProperty('IsAccessible')
    email = EConomicsProperty('Email')
    balance = EConomicsProperty('Balance')
    layout = EConomicsReference('Layout', 'TemplateCollection')
    vat_zone = EConomicsProperty('VatZone')
    debtor_contacts = EConomicsReferenceList('DebtorContacts', 'DebtorContact', 'Debtor_GetDebtorContacts')
    term_of_payment = EConomicsReference('TermOfPayment', 'TermOfPayment')
    quotations = EConomicsReferenceList('Quotations', 'Quotation', 'Debtor_GetQuotations')
    price_group = EConomicsReference('PriceGroup', 'PriceGroup')
    ean = EConomicsProperty('Ean')
    name = EConomicsProperty('Name')
    our_reference = EConomicsReference('OurReference', 'Employee')
    county = EConomicsProperty('County')
    ci_number = EConomicsProperty('CINumber')
    orders = EConomicsReferenceList('Orders', 'Order', 'Debtor_GetOrders')
    public_entry_number = EConomicsProperty('PublicEntryNumber')
    delivery_locations = EConomicsReferenceList('DeliveryLocations', 'DeliveryLocation', 'Debtor_GetDeliveryLocations')
    currency = EConomicsReference('Currency', 'Currency')
    current_invoices = EConomicsReferenceList('CurrentInvoices', 'CurrentInvoice', 'Debtor_GetCurrentInvoices')
    next_available_number = EConomicsProperty('NextAvailableNumber')
    vat_number = EConomicsProperty('VatNumber')
    debtor_group = EConomicsReference('DebtorGroup', 'DebtorGroup')
    credit_maximum = EConomicsProperty('CreditMaximum')
    entries = EConomicsReferenceList('Entries', 'Entry', 'Debtor_GetEntries')
    open_entries = EConomicsProperty('OpenEntries')
    subscribers = EConomicsReferenceList('Subscribers', 'Subscriber', 'Debtor_GetSubscribers')


class CreditorGroup(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'number', 'method': u'FindByNumber'}]
    creditors = EConomicsReferenceList('Creditors', 'Creditor', 'CreditorGroup_GetCreditors')
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')
    account = EConomicsReference('Account', 'Account')


class KeyFigureCode(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}]
    accounts = EConomicsReferenceList('Accounts', 'Account', 'KeyFigureCode_GetAccounts')
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class MileageEntry(EConomicsModel):
    __filters__ = []
    start_location = EConomicsProperty('StartLocation')
    id = EConomicsProperty('Id')
    end_location = EConomicsProperty('EndLocation')
    date = EConomicsProperty('Date')
    cost_rate = EConomicsProperty('CostRate')
    approved = EConomicsProperty('Approved')
    distance = EConomicsProperty('Distance')
    employee = EConomicsReference('Employee', 'Employee')
    sales_rate = EConomicsProperty('SalesRate')
    project = EConomicsReference('Project', 'Project')


class Project(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'number_list', 'method': u'FindByNumberList'}]
    debtor = EConomicsReference('Debtor', 'Debtor')
    entries = EConomicsReferenceList('Entries', 'Entry', 'Project_GetEntries')
    open_sub_projects = EConomicsProperty('OpenSubProjects')
    time_entries_by_date = EConomicsProperty('TimeEntriesByDate')
    project_group = EConomicsReference('ProjectGroup', 'ProjectGroup')
    is_main_project = EConomicsProperty('IsMainProject')
    entries_by_date = EConomicsProperty('EntriesByDate')
    description = EConomicsProperty('Description')
    is_closed = EConomicsProperty('IsClosed')
    name = EConomicsProperty('Name')
    main_project = EConomicsProperty('MainProject')
    activities = EConomicsReferenceList('Activities', 'Activity', 'Project_GetActivities')
    is_accessible = EConomicsProperty('IsAccessible')
    mileage_entries_by_date = EConomicsProperty('MileageEntriesByDate')
    time_entries = EConomicsReferenceList('TimeEntries', 'TimeEntry', 'Project_GetTimeEntries')
    number = EConomicsProperty('Number')
    responsible = EConomicsProperty('Responsible')
    mileage = EConomicsProperty('Mileage')


class Account(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    is_accessible = EConomicsProperty('IsAccessible')
    name = EConomicsProperty('Name')
    department = EConomicsReference('Department', 'Department')
    block_direct_entries = EConomicsProperty('BlockDirectEntries')
    sum_intervals = EConomicsReferenceList('SumIntervals', 'SumInterval', 'Account_GetSumIntervals')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    budget_figures_by_date = EConomicsProperty('BudgetFiguresByDate')
    balance = EConomicsProperty('Balance')
    budget_figures = EConomicsReferenceList('BudgetFigures', 'BudgetFigure', 'Account_GetBudgetFigures')
    type = EConomicsProperty('Type')
    number = EConomicsProperty('Number')
    vat_account = EConomicsReference('VatAccount', 'VatAccount')
    contra_account = EConomicsProperty('ContraAccount')
    entries_by_date = EConomicsProperty('EntriesByDate')
    opening_account = EConomicsProperty('OpeningAccount')
    entry_totals_by_date = EConomicsProperty('EntryTotalsByDate')
    debit_credit = EConomicsProperty('DebitCredit')
    entries_by_number = EConomicsProperty('EntriesByNumber')
    total_from = EConomicsProperty('TotalFrom')


class Activity(EConomicsModel):
    __filters__ = [{'name': u'number', 'method': u'FindByNumber'}, {'name': u'number_list', 'method': u'FindByNumberList'}]
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class Entry(EConomicsModel):
    __filters__ = [{'name': u'date_interval', 'method': u'FindByDateInterval'}, {'name': u'serial_number_interval', 'method': u'FindBySerialNumberInterval'}]
    account = EConomicsReference('Account', 'Account')
    project = EConomicsReference('Project', 'Project')
    currency = EConomicsReference('Currency', 'Currency')
    vat_account = EConomicsReference('VatAccount', 'VatAccount')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    amount_default_currency = EConomicsProperty('AmountDefaultCurrency')
    document = EConomicsProperty('Document')
    department = EConomicsReference('Department', 'Department')
    voucher_number = EConomicsProperty('VoucherNumber')
    last_used_serial_number = EConomicsProperty('LastUsedSerialNumber')
    type = EConomicsProperty('Type')
    text = EConomicsProperty('Text')
    amount = EConomicsProperty('Amount')
    serial_number = EConomicsProperty('SerialNumber')
    date = EConomicsProperty('Date')


class CurrentInvoiceLine(EConomicsModel):
    __filters__ = [{'name': u'product_list', 'method': u'FindByProductList'}, {'name': u'product', 'method': u'FindByProduct'}]
    total_net_amount = EConomicsProperty('TotalNetAmount')
    quantity = EConomicsProperty('Quantity')
    product = EConomicsReference('Product', 'Product')
    description = EConomicsProperty('Description')
    department = EConomicsReference('Department', 'Department')
    discount_as_percent = EConomicsProperty('DiscountAsPercent')
    unit_cost_price = EConomicsProperty('UnitCostPrice')
    margin_as_percent = EConomicsProperty('MarginAsPercent')
    unit_net_price = EConomicsProperty('UnitNetPrice')
    unit = EConomicsReference('Unit', 'Unit')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    invoice = EConomicsReference('Invoice', 'Invoice')
    delivery_date = EConomicsProperty('DeliveryDate')
    total_margin = EConomicsProperty('TotalMargin')
    number = EConomicsProperty('Number')


class SumInterval(EConomicsModel):
    __filters__ = []
    to_account = EConomicsProperty('ToAccount')
    from_account = EConomicsProperty('FromAccount')
    account = EConomicsReference('Account', 'Account')


class TermOfPayment(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}]
    debtor = EConomicsReference('Debtor', 'Debtor')
    distribution_in_percent2 = EConomicsProperty('DistributionInPercent2')
    contra_account2 = EConomicsProperty('ContraAccount2')
    name = EConomicsProperty('Name')
    days = EConomicsProperty('Days')
    distribution_in_percent = EConomicsProperty('DistributionInPercent')
    type = EConomicsProperty('Type')
    contra_account = EConomicsProperty('ContraAccount')
    description = EConomicsProperty('Description')


class TimeEntry(EConomicsModel):
    __filters__ = [{'name': u'all_updated', 'method': u'GetAllUpdated'}, {'name': u'all_updated', 'method': u'GetAllUpdated'}]
    text = EConomicsProperty('Text')
    approved = EConomicsProperty('Approved')
    date = EConomicsProperty('Date')
    project = EConomicsReference('Project', 'Project')
    number_of_hours = EConomicsProperty('NumberOfHours')
    employee = EConomicsReference('Employee', 'Employee')
    id = EConomicsProperty('Id')
    activity = EConomicsReference('Activity', 'Activity')


class PriceGroup(EConomicsModel):
    __filters__ = [{'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'name', 'method': u'FindByName'}]
    price = EConomicsProperty('Price')
    number = EConomicsProperty('Number')
    name = EConomicsProperty('Name')


class CreditorContact(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}]
    creditor = EConomicsReference('Creditor', 'Creditor')
    telephone_number = EConomicsProperty('TelephoneNumber')
    external_id = EConomicsProperty('ExternalId')
    comments = EConomicsProperty('Comments')
    email = EConomicsProperty('Email')
    number = EConomicsProperty('Number')
    name = EConomicsProperty('Name')


class DocumentArchiveCategory(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}]
    documents = EConomicsProperty('Documents')
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class VatAccount(EConomicsModel):
    __filters__ = [{'name': u'vat_code', 'method': u'FindByVatCode'}]
    rate_as_percent = EConomicsProperty('RateAsPercent')
    contra_account = EConomicsProperty('ContraAccount')
    type = EConomicsProperty('Type')
    account = EConomicsReference('Account', 'Account')
    name = EConomicsProperty('Name')
    vat_code = EConomicsProperty('VatCode')


class QuotationLine(EConomicsModel):
    __filters__ = [{'name': u'product_list', 'method': u'FindByProductList'}, {'name': u'product', 'method': u'FindByProduct'}]
    department = EConomicsReference('Department', 'Department')
    total_margin = EConomicsProperty('TotalMargin')
    number = EConomicsProperty('Number')
    total_net_amount = EConomicsProperty('TotalNetAmount')
    unit_net_price = EConomicsProperty('UnitNetPrice')
    margin_as_percent = EConomicsProperty('MarginAsPercent')
    description = EConomicsProperty('Description')
    discount_as_percent = EConomicsProperty('DiscountAsPercent')
    distribution_key = EConomicsReference('DistributionKey', 'DistributionKey')
    delivery_date = EConomicsProperty('DeliveryDate')
    quantity = EConomicsProperty('Quantity')
    unit = EConomicsReference('Unit', 'Unit')
    quotation = EConomicsReference('Quotation', 'Quotation')
    unit_cost_price = EConomicsProperty('UnitCostPrice')
    product = EConomicsReference('Product', 'Product')


class ScannedDocument(EConomicsModel):
    __filters__ = [{'name': u'voucher_number', 'method': u'FindByVoucherNumber'}, {'name': u'voucher_number_interval', 'method': u'FindByVoucherNumberInterval'}]
    to_date = EConomicsProperty('ToDate')
    from_date = EConomicsProperty('FromDate')
    page_count = EConomicsProperty('PageCount')
    status = EConomicsProperty('Status')
    number = EConomicsProperty('Number')
    voucher_number = EConomicsProperty('VoucherNumber')
    pdf = EConomicsProperty('Pdf')
    category = EConomicsProperty('Category')
    note = EConomicsProperty('Note')


class SubscriptionLine(EConomicsModel):
    __filters__ = [{'name': u'product_list', 'method': u'FindByProductList'}, {'name': u'subscripton_list', 'method': u'FindBySubscriptonList'}, {'name': u'product', 'method': u'FindByProduct'}, {'name': u'subscription', 'method': u'FindBySubscription'}]
    department = EConomicsReference('Department', 'Department')
    quantity = EConomicsProperty('Quantity')
    subscription = EConomicsReference('Subscription', 'Subscription')
    number = EConomicsProperty('Number')
    product = EConomicsReference('Product', 'Product')
    special_price = EConomicsProperty('SpecialPrice')
    product_name = EConomicsProperty('ProductName')


class DeliveryLocation(EConomicsModel):
    __filters__ = [{'name': u'external_id', 'method': u'FindByExternalId'}]
    address = EConomicsProperty('Address')
    postal_code = EConomicsProperty('PostalCode')
    terms_of_delivery = EConomicsProperty('TermsOfDelivery')
    debtor = EConomicsReference('Debtor', 'Debtor')
    country = EConomicsProperty('Country')
    city = EConomicsProperty('City')
    number = EConomicsProperty('Number')
    county = EConomicsProperty('County')
    is_accessible = EConomicsProperty('IsAccessible')
    external_id = EConomicsProperty('ExternalId')


class Invoice(EConomicsModel):
    __filters__ = [{'name': u'date_interval', 'method': u'FindByDateInterval'}, {'name': u'number_interval', 'method': u'FindByNumberInterval'}, {'name': u'our_reference', 'method': u'FindByOurReference'}, {'name': u'number', 'method': u'FindByNumber'}, {'name': u'order_number', 'method': u'FindByOrderNumber'}, {'name': u'other_reference', 'method': u'FindByOtherReference'}, {'name': u'number_list', 'method': u'FindByNumberList'}]
    vat_amount = EConomicsProperty('VatAmount')
    other_reference = EConomicsProperty('OtherReference')
    oio_xml = EConomicsProperty('OioXml')
    debtor_ean = EConomicsProperty('DebtorEan')
    debtor_postal_code = EConomicsProperty('DebtorPostalCode')
    terms_of_delivery = EConomicsProperty('TermsOfDelivery')
    remainder = EConomicsProperty('Remainder')
    delivery_county = EConomicsProperty('DeliveryCounty')
    our_reference = EConomicsReference('OurReference', 'Employee')
    currency = EConomicsReference('Currency', 'Currency')
    debtor_name = EConomicsProperty('DebtorName')
    date = EConomicsProperty('Date')
    net_amount = EConomicsProperty('NetAmount')
    number = EConomicsProperty('Number')
    remainder_default_currency = EConomicsProperty('RemainderDefaultCurrency')
    order_number = EConomicsProperty('OrderNumber')
    term_of_payment = EConomicsReference('TermOfPayment', 'TermOfPayment')
    delivery_postal_code = EConomicsProperty('DeliveryPostalCode')
    debtor_address = EConomicsProperty('DebtorAddress')
    project = EConomicsReference('Project', 'Project')
    attention = EConomicsProperty('Attention')
    is_vat_included = EConomicsProperty('IsVatIncluded')
    due_date = EConomicsProperty('DueDate')
    our_reference2 = EConomicsReference('OurReference2', 'Employee')
    delivery_date = EConomicsProperty('DeliveryDate')
    public_entry_number = EConomicsProperty('PublicEntryNumber')
    debtor_city = EConomicsProperty('DebtorCity')
    layout = EConomicsReference('Layout', 'TemplateCollection')
    text_line1 = EConomicsProperty('TextLine1')
    text_line2 = EConomicsProperty('TextLine2')
    rounding_amount = EConomicsProperty('RoundingAmount')
    pdf = EConomicsFileProperty('Pdf', 'Invoice_GetPdf', 'pdf')
    debtor_country = EConomicsProperty('DebtorCountry')
    delivery_location = EConomicsReference('DeliveryLocation', 'DeliveryLocation')
    heading = EConomicsProperty('Heading')
    delivery_city = EConomicsProperty('DeliveryCity')
    debtor = EConomicsReference('Debtor', 'Debtor')
    gross_amount = EConomicsProperty('GrossAmount')
    lines = EConomicsReferenceList('Lines', 'InvoiceLine', 'Invoice_GetLines')
    your_reference = EConomicsProperty('YourReference')
    deduction_amount = EConomicsProperty('DeductionAmount')
    net_amount_default_currency = EConomicsProperty('NetAmountDefaultCurrency')
    debtor_county = EConomicsProperty('DebtorCounty')
    delivery_country = EConomicsProperty('DeliveryCountry')
    delivery_address = EConomicsProperty('DeliveryAddress')


class DebtorContact(EConomicsModel):
    __filters__ = [{'name': u'external_id', 'method': u'FindByExternalId'}, {'name': u'name', 'method': u'FindByName'}]
    is_to_receive_email_copy_of_order = EConomicsProperty('IsToReceiveEmailCopyOfOrder')
    comments = EConomicsProperty('Comments')
    debtor = EConomicsReference('Debtor', 'Debtor')
    number = EConomicsProperty('Number')
    email = EConomicsProperty('Email')
    name = EConomicsProperty('Name')
    is_to_receive_email_copy_of_invoice = EConomicsProperty('IsToReceiveEmailCopyOfInvoice')
    telephone_number = EConomicsProperty('TelephoneNumber')
    external_id = EConomicsProperty('ExternalId')


class ProjectGroup(EConomicsModel):
    __filters__ = [{'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}]
    projects = EConomicsReferenceList('Projects', 'Project', 'ProjectGroup_GetProjects')
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


class Subscription(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}]
    number = EConomicsProperty('Number')
    description = EConomicsProperty('Description')
    collection = EConomicsProperty('Collection')
    calendar_year_basis = EConomicsProperty('CalendarYearBasis')
    include_name = EConomicsProperty('IncludeName')
    subscribers = EConomicsReferenceList('Subscribers', 'Subscriber', 'Subscription_GetSubscribers')
    next_available_number = EConomicsProperty('NextAvailableNumber')
    subscription_interval = EConomicsProperty('SubscriptionInterval')
    subscription_lines = EConomicsReferenceList('SubscriptionLines', 'SubscriptionLine', 'Subscription_GetSubscriptionLines')
    allow_more_than_one_for_each_debtor = EConomicsProperty('AllowMoreThanOneForEachDebtor')
    add_period = EConomicsProperty('AddPeriod')
    name = EConomicsProperty('Name')


class CashBook(EConomicsModel):
    __filters__ = [{'name': u'name', 'method': u'FindByName'}, {'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'name_list', 'method': u'FindByNameList'}, {'name': u'number', 'method': u'FindByNumber'}]
    number = EConomicsProperty('Number')
    entries = EConomicsReferenceList('Entries', 'Entry', 'CashBook_GetEntries')
    next_voucher_number = EConomicsProperty('NextVoucherNumber')
    name = EConomicsProperty('Name')


class Unit(EConomicsModel):
    __filters__ = [{'name': u'number_list', 'method': u'FindByNumberList'}, {'name': u'name', 'method': u'FindByName'}, {'name': u'number', 'method': u'FindByNumber'}]
    name = EConomicsProperty('Name')
    number = EConomicsProperty('Number')


#!/usr/bin/env python

from distutils.core import setup

with open('README.md') as f:
    docs = f.read()

setup(
    name='pyconomic',
    version='0.1.1',
    description='Abstraction for e-conomic.com API',
    author='Mikkel Jans',
    author_email='mikkeljans@gmail.com',
    url='https://github.com/mikkeljans/pyconomic',
    packages=['pyconomic'],
    long_description=docs
)


"""
Model Abstraction of e-economic.com API
"""

import copy
import re
import os
import base64
from collections import defaultdict

from suds.client import Client


class ObjectDoesNotExist(BaseException):
    pass


class MultipleObjectsReturned(BaseException):
    pass


class EConomicsService(object):
    """
    Interface for e-conomic WSDL service
    """
    def __init__(self, service, model_factory, soap_factory, codec):
        self.service = service
        self.model_factory = model_factory
        self.soap_factory = soap_factory
        self.ncalls = 0
        self.codec = codec

    def fetch_list(self, name, expected_wsdltype, *args, **kw):
        result = getattr(self.service, name)(*args)
        self.ncalls += 1
        if not result:
            return []
        if expected_wsdltype and expected_wsdltype not in result.__keylist__:
            return [result]
        return result[0]

    def fetch(self, name, *args, **kw):
        return getattr(self.service, name)(*args)

    def upgrade_to_order(self, handle, order_model):
        hnd = self.fetch('Quotation_UpgradeToOrder', handle)
        return self.model_factory.get_or_create_instance(self, order_model, hnd)

    def upgrade_to_invoice(self, handle, current_invoice_model):
        hnd = self.fetch('Order_UpgradeToInvoice', handle)
        return self.model_factory.get_or_create_instance(self, current_invoice_model, hnd)

    def book_invoice(self, handle, invoice_model):
        hnd = self.fetch('CurrentInvoice_Book', handle)
        return self.model_factory.get_or_create_instance(self, invoice_model, hnd)

    def next_available_number(self, model):
        return self.fetch('%s_GetNextAvailableNumber' % model.__name__)

    def delete(self, model, handle):
        self.fetch("%s_Delete" % model.__name__, handle)

    def create(self, model, **data):
        parsed_data = self.codec.encode_data_object(self, model, data)
        hnd = self.fetch("%s_CreateFromData" % model.__name__, parsed_data)
        return self.get_instance(model, hnd)

    def get_or_create(self, model, **spec):
        filter_names = [f['name'] for f in model.__filters__]
        get_data = dict((k, v,) for k, v in spec.items() if k in filter_names)
        try:
            return self.get(model, **get_data)
        except ObjectDoesNotExist:
            return self.create(model, **spec)

    def __find_handles(self, model, **spec):
        """ find model instances based on given filter (spec)
        The filter is based on available server-calls, so some values might not be available for filtering.
        Multiple filter-values is going to do multiple server-calls.
        For complex filters in small datasets, it might be faster to fetch all and do your own in-memory filter.
        Empty filter will fetch all.

        :param model: subclass of EConomicsModel
        :param spec: mapping of values to filter by
        :return: a list of EConomicsModel instances
        """
        server_calls = []
        filter_names = dict([(f['name'], f['method'],) for f in model.get_filters()])
        if not spec:
            server_calls.append({'method': "%s_GetAll" % model.__name__, 'args': []})
        else:
            for key, value in spec.items():
                if not key in filter_names:
                    raise ValueError("no server-method exists for filtering by '%s'" % key)
                args = []
                if not hasattr(value, '__iter__'):
                    value = [value]

                if key.endswith('_list'):
                    vtype = type(value[0]).__name__
                    # TODO: this surely does not cover all cases of data types
                    array = self.soap_factory.create('ArrayOf%s' % vtype.capitalize())
                    getattr(array, "%s" % vtype).extend(value)
                    args.append(array)
                else:
                    args.extend(value)

                method = "%s_%s" % (model.__name__, filter_names[key])
                if filter_names[key].startswith('GetAll'):
                    args = []
                server_calls.append({'method': method, 'args': args, 'expect': "%sHandle" % model.__name__})

        handles = [
            map(Handle, self.fetch_list(scall['method'], scall.get('expect'), *scall['args']))
            for scall in server_calls
        ]
        return [h.wsdl for h in reduce(set.intersection, map(set, handles))]

    def find(self, model, **spec):
        handles = self.__find_handles(model, **spec)
        return [self.get_instance(model, hnd) for hnd in handles]

    def get(self, model, **spec):
        """get a single model instance by handle

        :param model: model
        :param handle: instance handle
        :return:
        """

        handles = self.__find_handles(model, **spec)
        if len(handles) > 1:
            raise MultipleObjectsReturned()
        if not handles:
            raise ObjectDoesNotExist()
        return self.get_instance(model, handles[0])

    def get_instance(self, model, handle):
        return self.model_factory.get_or_create_instance(self, model, handle)

    def load_instance_data(self, instance):
        model = instance.__class__
        modelname = model.__name__
        data = self.fetch("%s_GetData" % modelname, instance._handle)
        instance._data = self.codec.decode_data_object(self, instance._handle, model, data)

    def load_data(self, instance):
        model = instance.__class__
        modelname = model.__name__
        handles = [inst._handle for (m, inst,) in self.model_factory.instances_iter([model], loaded=False)]
        array = self.soap_factory.create('ArrayOf%sHandle' % modelname)
        getattr(array, "%sHandle" % modelname).extend(handles)

        for data in self.fetch_list("%s_GetDataArray" % modelname, None, array):
            handle = data.Handle
            inst = self.get_instance(model, handle)
            inst._data = self.codec.decode_data_object(self, handle, model, data)
            inst._loaded = True

    def get_all_changes(self):
        changesets = defaultdict(list)
        for model, inst in self.model_factory.instances_iter(updated=True):
            changesets[model].append(ModelChange(model, inst))
        return changesets

    def commit(self):
        changesets = self.get_all_changes()
        for model, changes in changesets.items():
            datalist = [self.codec.encode_data_object(self, model, changeset.get_data()) for changeset in changes]
            array = self.soap_factory.create('ArrayOf%sData' % model.__name__)
            getattr(array, '%sData' % model.__name__).extend(datalist)

            self.fetch("%s_UpdateFromDataArray" % model.__name__, array)
            [change.apply_and_clear() for change in changes]

    def __getattr__(self, name):
        return getattr(self.service, name)


class ModelChange(object):
    def __init__(self, model, instance):
        self.model = model
        self.instance = instance

    def __repr__(self):
        return "<Changes %r %r>" % (self.instance, self.clean_data(self.instance._changes))

    def apply_and_clear(self):
        self.instance._data.update(self.instance._changes)
        self.instance._changes = {}

    def clean_data(self, data):
        result = {}
        for k, v in data.items():
            k = pythonize(k)
            if k.endswith('_handle'):
                k = k[:-7]
            result[k] = v
        return result

    def get_data(self):
        if not self.instance._data:
            self.instance.fetch()
        data = self.clean_data(self.instance._data)
        data.update(self.clean_data(self.instance._changes))
        data['Handle'] = self.instance._handle
        return data


class PropertyCodec(object):
    def __init__(self, missing_value=None):
        self.missing_value = missing_value

    def decode_data_object(self, service, handle, model, data):
        decoded_data = {}
        for prop in model.properties:
            name = prop.name
            if prop.name+'Handle' in data:
                name = prop.name + 'Handle'
            if not name in data:
                value = prop.default_value(service, handle)
            else:
                value = prop.decode_value(service, handle, data[name])
            decoded_data[prop.name] = value
        return decoded_data

    def encode_data_object(self, service, model, data):
        #print 'ENCODE', data
        encoded_data = {}
        if 'Handle' in data:
            encoded_data['Handle'] = data['Handle']
        for prop in model.properties:
            name = prop.pyname
            if not name in data:
            #    encoded_data[prop.name] = self.missing_value
                continue
            value = data[name]
            if value is None:
            #    encoded_data[prop.name] = value
                continue
            encoded_data[prop.name] = prop.encode_value(service, data[name])
        return encoded_data


class EConomicsModelFactory(object):
    def __init__(self):
        self.__models = {}

    def instances_iter(self, models=None, loaded=None, updated=None):
        if models is None:
            models = self.__models.keys()
        for model in models:
            for inst in self.__models[model].values():
                if loaded is not None and bool(inst._loaded) != bool(loaded):
                    continue
                if updated is not None and bool(inst._changes) != bool(updated):
                    continue
                yield (model, inst,)

    def get_or_create_instance(self, service, model, handle):
        hashkey = hash((service, model, handle[0],))
        modeldata = self.__models.setdefault(model, {})
        return modeldata.setdefault(hashkey, model(service, handle))


class Handle(object):
    def __init__(self, wsdl):
        self.wsdl = wsdl

    def __hash__(self):
        return hash(self.wsdl[0])

    def __eq__(self, other):
        return hash(self) == other

    def __repr__(self):
        return "<Handle %r>" % self.wsdl.Id


class EConomicsMeta(type):
    registry = {}

    def __new__(mcs, name, bases, ns):
        properties = []
        for k, v in ns.items():
            if hasattr(v, '__get__'):
                properties.append(v)
        ns['properties'] = properties
        model = type.__new__(mcs, name, bases, ns)

        mcs.registry[name] = model
        return model

    def get_filters(self):
        return self.__filters__


class EConomicsBaseProperty(object):
    def encode_value(self, service, value):
        return value

    def decode_value(self, service, handle, value):
        return value

    def default_value(self, service, handle):
        return None

    def __get__(self, instance, owner):
        _ = owner
        if instance is None:
            return self

        changes = instance._changes
        if self.name in changes:
            return changes[self.name]

        if not instance._loaded:
            instance.load()

        value = instance._data[self.name]
        if hasattr(value, 'fetched') and not value.fetched:
            value.fetch()

        return value

    def __set__(self, instance, value):
        instance._changes[self.name] = value


class EConomicsProperty(EConomicsBaseProperty):
    def __init__(self, name):
        self.name = name
        self.pyname = pythonize(name)

    def __repr__(self):
        return "<%s Data>" % pythonize(self.name)


class EConomicsReference(EConomicsBaseProperty):
    def __init__(self, name, model):
        self.name = name + 'Handle'
        self.model = model
        self.pyname = pythonize(name)

    def encode_value(self, service, value):
        return value._handle

    def decode_value(self, service, handle, value):
        return service.get_instance(get_model(self.model), value)

    def __repr__(self):
        return "<%s %s>" % (self.name, self.model)


class QueryList(list):
    def __init__(self, service, handle, model, method):
        self.service = service
        self.handle = handle
        self.model = model
        self.method = method
        self.fetched = False

    def __getattribute__(self, name):
        if name in ['fetch', 'service', 'handle', 'model', 'method', 'fetched']:
            return list.__getattribute__(self, name)
        if self.fetched:
            self.fetch()
        return list.__getattribute__(self, name)

    def fetch(self):
        handles = self.service.fetch_list(self.method, None, self.handle)
        self[:] = [self.service.get_instance(self.model, hnd) for hnd in handles]
        self.fetched = True
        return self


class EConomicsReferenceList(EConomicsBaseProperty):
    def __init__(self, name, model, method):
        self.name = name
        self.model = model
        self.method = method
        self.pyname = pythonize(name)

    def __repr__(self):
        return "<%s [%s]>" % (self.name, self.model)

    def encode_value(self, service, value):
        return [v._handle for v in value]

    def default_value(self, service, handle):
        return QueryList(service, handle, get_model(self.model), self.method)


class EConomicsFileProperty(EConomicsBaseProperty):
    def __init__(self, name, method, filetype):
        self.name = name
        self.filetype = filetype
        self.method = method
        self.pyname = pythonize(name)

    def __repr__(self):
        return "<%s %s file>" % (self.name, self.filetype)

    def default_value(self, service, handle):
        return FileObject(service, self.method, handle, self.filetype)


class FileObject(object):
    def __init__(self, service, method, handle, filetype):
        self.filedata = None
        self.method = method
        self.service = service
        self.handle = handle
        self.filetype = filetype
        self.fetched = False
        self.__last_location = None

    def fetch(self):
        self.filedata = self.service.fetch(self.method, self.handle)
        self.fetched = True
        return self

    def save(self, location):
        if not location.endswith(self.filetype):
            location += '.' + self.filetype
        with open(location, 'wb') as f:
            f.write(base64.b64decode(self.filedata))
        self.__last_location = location

    def show(self):
        if not self.__last_location:
            self.save('/tmp/economic_tmp')
        os.system('xdg-open %s' % self.__last_location)


class EConomicsModel(object):
    __filters__ = []
    __metaclass__ = EConomicsMeta

    def __init__(self, service, handle):
        self._handle = handle
        self._loaded = False
        self._service = service
        self._data = {}
        self._changes = {}

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self._handle[0])

    def fetch(self):
        self._service.load_instance_data(self)
        return self

    def update(self, **data):
        for k, v in data.items():
            setattr(self, k, v)

    def load(self):
        self._service.load_data(self)

    def delete(self):
        self._service.delete(self.__class__, self._handle)


def get_model(name):
    return EConomicsMeta.registry[name]


def pythonize(name):
    return re.sub('([A-Z])([a-z])', r'_\1\2', name).strip('_').lower()


def camelcase(name):
    return ''.join(map(str.capitalize, name.split('_')))


def build_model_code(client):
    """
    Generate source code for e-conomic models based on WSDL connection.
    This is based on the assumption that the API follows a specific method naming-convention.
    Not all models and attributes has been tested.
    The source-generation is mostly to help improve readability and IDE auto-completion.

    :param client:
    :return: source code for models.py
    """
    models = {}
    references = {}
    for method in client.wsdl.services[0].ports[0].methods.values():
        if not '_' in method.name:
            continue
        model, action = method.name.split('_')
        models.setdefault(model, {'properties': [], 'filters': []})
        references[model] = model
        if model[-1] == 'y':
            references[model[:-1] + 'ies'] = model
        else:
            references[model+'s'] = model

    references['OurReference'] = 'Employee'
    references['GetYourReference'] = 'DebtorContact'
    references['GetAttention'] = 'DebtorContact'
    references['Layout'] = 'TemplateCollection'
    special = {
        'Order_GetPdf': {
            'type': 'EConomicsFileProperty',
            'args': ["'Order_GetPdf'", "'pdf'"]
        },
        'Invoice_GetPdf': {
            'type': 'EConomicsFileProperty',
            'args': ["'Invoice_GetPdf'", "'pdf'"]
        },
        'CurrentInvoice_GetPdf': {
            'type': 'EConomicsFileProperty',
            'args': ["'CurrentInvoice_GetPdf'", "'pdf'"]
        }
    }
    for line in ['Order', 'Invoice', 'CurrentInvoice', 'Quotation']:
        method = '%s_GetLines' % line
        special[method] = {
            'type': 'EConomicsReferenceList',
            'args': ["'%sLine'" % line, "'%s'" % method]
        }

    for method in client.wsdl.services[0].ports[0].methods.values():
        if not '_' in method.name:
            continue
        model, action = method.name.split('_')
        if action in ['GetData', 'GetAll', 'GetDataArray']:
            continue
        modeldata = models[model]

        if action == 'GetAllUpdated':
            camelname = action[3:]
            modeldata['filters'].append({'name': pythonize(camelname), 'method': action})
        if re.findall('GetAll[A-Z].+', action):
            camelname = action[3:]
            modeldata['filters'].append({'name': pythonize(camelname), 'method': action})
        elif action.startswith('FindBy'):
            camelname = action[6:]
            modeldata['filters'].append({'name': pythonize(camelname), 'method': action})

        elif action.startswith('Get'):
            propname = action[3:]
            pyname = pythonize(propname)
            if not propname:
                continue
            get_type = re.findall('Get(%s)[a-z0-9]*?$' % ('|'.join(references.keys())), action)
            if get_type and get_type[0] in references:
                refmodel = references[get_type[0]]
                if action[-1] == 's':
                    modeldata['properties'].append({
                        'type': 'EConomicsReferenceList',
                        'args': ["'%s'" % propname, "'%s'" % refmodel,  "'%s'" % method.name],
                        'name': pyname
                    })
                else:
                    modeldata['properties'].append({
                        'type': 'EConomicsReference',
                        'args': ["'%s'" % propname, "'%s'" % refmodel],
                        'name': pyname
                    })
            elif method.name in special:
                spdata = special[method.name]
                modeldata['properties'].append({
                    'type': spdata['type'],
                    'args': ["'%s'" % propname] + spdata['args'],
                    'name': pyname
                })
            else:
                modeldata['properties'].append({
                    'type': 'EConomicsProperty',
                    'args': ["'%s'" % propname],
                    'name': pyname
                })

    classes = []
    for modelname, modeldata in models.items():

        propertycode = ["%s = %s(%s)" % (md['name'], md['type'], ', '.join(md['args']))
                        for md in modeldata['properties']]
        code = "class %s(%s):\n    __filters__ = %r\n    %s" % (modelname, 'EConomicsModel',
                                                                modeldata['filters'], '\n    '.join(propertycode))
        classes.append(code)
    return "from pyconomic.base import *\n\n\n" + "\n\n\n".join(classes)
